const char *rprf_rcsid = "$Header: /u/spert/src/rprf/RCS/rprf.c,v 1.24 1996/10/16 19:28:51 krste Exp $";
/*
   Routines to help with performance testing.
*/

#include <stdarg.h>
#include <stdlib.h>
#include <float.h>
#include <stdio.h>
#include <string.h>

/* For systems with broken header files. */
/* FIXME: Should be handled by configure. */
int fprintf(FILE*, const char*, ...);
int printf(const char*, ...);
long strtol(const char*, char**, int);
double strtod(const char*, char**);

#include "rprf.h"

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#endif
#ifndef EXIT_FAILURE
#define EXIT_FAILURE (1)
#endif

static void
error(const char *msg)
{
    fprintf(stderr, "rprf: %s\n", msg);
    exit(EXIT_FAILURE);
}

static const char* progname;	/* Name of the current perf program. */
static const char *user_format; /* Format string user provided. */

/*****************************************************************************/
/*
  Variables dealing with timing loops.  Code is in header file.
*/
/*****************************************************************************/

/* Default number of repeats. */
#ifndef RPRF_N_INNER_DEFAULT
# define RPRF_N_INNER_DEFAULT (10L)
#endif
long rprf_n_inner = RPRF_N_INNER_DEFAULT;

/* Default number of outer loops. */
#ifndef RPRF_N_OUTER_DEFAULT
# define RPRF_N_OUTER_DEFAULT  (10L)
#endif
long rprf_n_outer = RPRF_N_OUTER_DEFAULT;

enum rprf_R_t rprf_R = RPRF_R_NONE;
enum rprf_S_t rprf_S = RPRF_S_MAX_RATE;

int rprf_verbose = 0;           /* Be verbose? */
int rprf_i_auto = 0;            /* Use automatic inner loop count? */

/* Minimum time for auto inner loop. */
#ifndef RPRF_I_MIN_T_DEFAULT
# define RPRF_I_MIN_T_DEFAULT (0.0)
#endif
double rprf_i_min_t = RPRF_I_MIN_T_DEFAULT;

/* Minimum resolution multiplier. */
#ifndef RPRF_I_MIN_M_DEFAULT
# define RPRF_I_MIN_M_DEFAULT (1000L)
#endif
long rprf_i_min_m = RPRF_I_MIN_M_DEFAULT;

/*****************************************************************************/
/*
  Code dealing with memory arena.
*/
/*****************************************************************************/

/* 16 MB standard arena size. */
#ifndef ARENA_LG2SIZE_DEFAULT
# define ARENA_LG2SIZE_DEFAULT (24L)
#endif

static long arena_lg2size = ARENA_LG2SIZE_DEFAULT;

/* 8 KB standard cache size, works as page alignment on most systems. */
#ifndef CACHE_LG2SIZE_DEFAULT
#  define CACHE_LG2SIZE_DEFAULT (13L)
#endif

static long cache_lg2size = CACHE_LG2SIZE_DEFAULT;

static char *arenap = NULL;       /* Pointer to memory arena. */
static size_t arena_next = 0;     /* Start of next place to allocate. */
static size_t arena_size = 0;     /* Size of arena. */

/* FIXME: Only use simple table to hold list of offsets. */
#ifndef MAX_OFFSETS
#  define MAX_OFFSETS (64)
#endif

static int n_offsets = 0;
static int next_offset = 0;
static long offset[MAX_OFFSETS];

static void
arena_init(void)
{
    /* Work out arena start alignment. */
    const size_t cache_size = 1 << cache_lg2size;
    const size_t cache_mask = cache_size - 1;

    /* Initialize memory arena. */
    arena_next = 0;
    arena_size = (size_t) (1 << arena_lg2size);

    arenap = (char *) malloc(arena_size + cache_mask);
    if (arenap==NULL)
        error("Couldn't allocate memory arena.");

#ifdef DEBUG_PRINTF
    fprintf(stderr, "Original arenap=%08lx, ", (unsigned long) arenap);
#endif

    /* Bump arenap to start on cache-size-aligned boundary. */
    /* FIXME: Casting pointer to unsigned long!!! */
    arenap = (char *) (((unsigned long) arenap + cache_mask) & ~ cache_mask);

#ifdef DEBUG_PRINTF
    fprintf(stderr, "aligned arenap=%08lx\n", (unsigned long) arenap);
#endif
}

void *
rprf_malloc(size_t len, size_t alignment)
{
    const size_t align_mask = alignment - 1;
    void * retp;                  /* Value to return. */

    /* Bump index according to offset info. */
    if (next_offset < n_offsets)
    {
        /* First move end to next cache boundary. */
        const size_t cache_size = 1 << cache_lg2size;
        const size_t cache_mask = cache_size - 1;
        arena_next = (arena_next + cache_mask) & ~cache_mask;

        /* Then add on offset. */
        arena_next += offset[next_offset++];
    }
    /* Else abut next block. */

    /* Bump index to next aligned boundary. */
    if (alignment & align_mask)
        error("rprf_malloc(): Alignment not a power of 2.");
    arena_next = (arena_next + align_mask) & ~align_mask;

    /* Store pointer to return for this malloc. */
    retp = (void *) (arenap + arena_next);

    /* Bump index past end. */
    arena_next += len;

    if (arena_next > arena_size)
        error("rprf_malloc(): "
              "No more space in memory arena, increase -A argument.");

#ifdef DEBUG_PRINTF
    fprintf(stderr, "retp=%08lx\n", (unsigned long) retp);
#endif

    return retp;
}

/*****************************************************************************/
/*
  Code to find timer overhead.
*/
/*****************************************************************************/

int rprf_t_comp = 0;            /* Compensate for timer call overhead. */
double rprf_t_overhead = 0.0;

static void
find_timer_overhead(void)
{
    if (rprf_t_comp)
    {
        double min_overhead = 1e30;
        long i;

        if (rprf_verbose)
        {
            printf("Determining timer call overhead.\n");
            printf("Measuring time for empty calls.\n");
        }

        for (i = 0; i < rprf_n_outer; i++)
        {
            IPM_timer timer;
            double t;

            IPM_timer_clear(&timer);
            IPM_timer_start(&timer);
            IPM_timer_stop(&timer);

            t = IPM_timer_read(&timer);
            if (rprf_verbose)
                printf(" %gs\n", t);

            if (t < min_overhead)
                min_overhead = t;
        }

        if (rprf_verbose)
        {
            printf("Using minimum observed overhead of %gs.\n", min_overhead);
        }

        rprf_t_overhead = min_overhead;
    }
}

/*****************************************************************************/
/* Argument parsing code. */
/*****************************************************************************/

static void
parse_long(const char*const s, long *const lp)
{
    size_t len = strlen(s);
    char *ptr;

    *lp = strtol(s, &ptr, 0);

    if (ptr != (s+len))
        error("Not an integer argument.");
}

static void
parse_float(const char*const s, float *const fp)
{
    size_t len = strlen(s);
    char *ptr;
    double val;
    double absval;

    val = strtod(s, &ptr);

    if (ptr != (s+len))
        error("Not a floating point argument.");

    /* Do fabs() here to avoid linking with math library. */
    absval = (val < 0) ? -val : val;

    if ( (absval != 0.0 && absval < FLT_MIN) || absval > FLT_MAX)
        error("Float argument out of range.");

    *fp = (float) val;
}

static void
parse_double(const char*const s, double *const dp)
{
    size_t len = strlen(s);
    double val;
    char *ptr;

    val = strtod(s, &ptr);

    if (ptr != (s+len))
        error("Not a floating point argument.");

    if (val < -DBL_MAX || val > DBL_MAX)
        error("Float argument out of range.");
    *dp = val;
}

static void
usage(void)
{
    fprintf(stderr, "Usage: %s [<options>] [<perf_args>]\n"
"where <options> is zero of more of:\n"
"  -help                Print this message.\n"
"  -i <inner_count=%ld> Use fixed inner loop count.\n"
"  -i-auto              Use automatically determined inner loop count.\n"
"  -i-min-t <time=%g>   With -i-auto, set minimum inner loop time in seconds.\n"
"  -i-min-m <multiplier=%ld> With -i-auto, inner loop time at least this factor\n"
"                            greater than minimum timer resolution.\n"
"  -t-comp              Try to compensate for timer call overhead.\n"
"  -o <outer_count=%ld> Use fixed outer loop count.\n"
"  -A <log2_size_of_arena=%ld> Log base 2 of size of memory arena.\n"
"  -c <log2_size_of_cache=%ld> Log base 2 of size of cache.\n"
"  {-a <next_alignment_offset=0>}  Alignment of next vector.\n"
"Only last specified -R or -S option for results display is used:\n"
"  -R  (deprecated - use -R-per-outer)\n"
"  -R-per-outer     For each outer loop, show total inner time.\n"
"  -R-per-inner     For each outer loop, show time for single inner iteration.\n"
"  -R-rate          For each outer loop, show Mops rate.\n"
"  -S-max-rate      Show maximum rate (default).\n"
"  -S-min-per-inner Show minimum time for single inner iteration.\n"
"  -v                   Turn on verbose messages about timing process.\n"
"and <perf_args> follow the format string\n"
"  %s\n"
"where\n"
"  i represents an integer\n"
"  f represents a single-precision floating-point number\n"
"  d represents a double-precision floating-point number\n"
"  s represents a string\n",
	    progname,
            RPRF_N_INNER_DEFAULT,
            RPRF_I_MIN_T_DEFAULT,
            RPRF_I_MIN_M_DEFAULT,
            RPRF_N_OUTER_DEFAULT,
            ARENA_LG2SIZE_DEFAULT,
            CACHE_LG2SIZE_DEFAULT,
            user_format);
}

static void
usage_error(const char *msg)
{
    fprintf(stderr, "Error: %s\n", msg);
    usage();
    exit(EXIT_FAILURE);
}

void
rprf_parse_args(int argc, char* argv[], const char* format, ... )
{
    va_list ap;

    /* Never look at argv[0]. */
    argc--;
    progname = *argv++;
    user_format = format;

    /* First parse standard rprf arguments. */

    while (argc)
    {
        const char *argp = *argv;

        if (argp[0] == '-')
        {
            /* Move to next argument. */
            argc--;
            argv++;

            switch (argp[1])
            {
            case 'h':
                /* Show help message. */
                usage();
                exit(EXIT_SUCCESS);
                break;

            case 'i':
                if (strcmp(argp, "-i")==0)
                {
                    /* Number of inner timing loops. */
                    if (argc--)
                        parse_long(*argv++, &rprf_n_inner);
                    else
                        usage_error("Argument to -i missing.");
                }
                else if (strcmp(argp, "-i-auto")==0)
                {
                    /* Use automatic inner loop count. */
                    rprf_i_auto = 1;
                }
                else if (strcmp(argp, "-i-min-t")==0)
                {
                    /* Set minimum time of inner loop. */
                    if (argc--)
                        parse_double(*argv++, &rprf_i_min_t);
                    else
                        usage_error("Argument to -i-min-t missing.");
                }
                else if (strcmp(argp, "-i-min-m")==0)
                {
                    /* Set minimum multiplier of clock resolution. */
                    if (argc--)
                        parse_long(*argv++, &rprf_i_min_m);
                    else
                        usage_error("Argument to -i-min-m missing.");
                }
                else
                    usage_error("Invalid argument.");
                break;

            case 'o':
                /* Number of outer timing loops. */
                if (argc--)
                    parse_long(*argv++, &rprf_n_outer);
                else
                    usage_error("Argument to -o missing.");
                break;

            case 't':
                if (strcmp(argp,"-t-comp")==0)
                {
                    /* Compensate for timer overhead. */
                    rprf_t_comp = 1;
                }
                else
                    usage_error("Invalid argument.");
                break;

            case 'A':
                /* Size of memory arena, as power of 2 number of bytes. */
                if (argc--)
                {
                    parse_long(*argv++, &arena_lg2size);

                    if (arena_lg2size < 1 || arena_lg2size > 31)
                        usage_error(
                            "Memory arena size negative or too large.");
                }
                else
                    usage_error("Argument to -A missing.");
                break;

            case 'c':
                /* Size of cache, as power of 2 number of bytes. */
                if (argc--)
                    parse_long(*argv++, &cache_lg2size);
                else
                    usage_error("Argument to -c missing.");
                break;

            case 'a':
                /* Size of next alignment offset, given in bytes. */
                if (argc--)
                {
                    if (n_offsets==MAX_OFFSETS)
                        error("Too many alignment offsets.");

                    parse_long(*argv++, &offset[n_offsets++]);
                }
                else
                    usage_error("Argument to -a missing.");
                break;

            case 'R':
                if (strcmp(argp, "-R")==0 || strcmp(argp, "-R-per-outer")==0)
                {
                    rprf_R = RPRF_R_PER_OUTER;
                }
                else if (strcmp(argp, "-R-per-inner")==0)
                {
                    rprf_R = RPRF_R_PER_INNER;
                }
                else if (strcmp(argp, "-R-rate")==0)
                {
                    rprf_R = RPRF_R_RATE;
                }
                else
                    usage_error("Invalid argument");

                rprf_S = RPRF_S_NONE; /* Turn off summary times. */
                break;

            case 'S':
                if (strcmp(argp, "-S-max-rate")==0)
                {
                    rprf_S = RPRF_S_MAX_RATE;
                }
                else if (strcmp(argp, "-S-min-per-inner")==0)
                {
                    rprf_S = RPRF_S_MIN_PER_INNER;
                }
                else
                    usage_error("Invalid argument.");

                rprf_R = RPRF_R_NONE; /* Turn off raw times. */
                break;

            case 'v':
                /* Turn on verbose messages. */
                rprf_verbose = 1;
                break;

            default:
                usage_error("Invalid argument.");
		break;
            }
        }
        else
            break;              /* Switch to format string arguments. */
    }

    if (rprf_verbose)
    {
        /* FIXME: Add better argument parsing status report. */
        printf("All timings using %ld outer loops.\n", rprf_n_outer);
    }

    /* Now parse any remaining arguments according to format string. */
    va_start(ap, format);

    if (argc < strlen(format))
        usage_error("Not enough <perf_args>.");

    while (argc--)
    {
        long *lp;
        float *fp;
        double *dp;
	const char** sp;

        if (rprf_S!=RPRF_S_NONE)
            printf("%s ", *argv);

        switch (*format++)
        {
        case 'i':
            lp = va_arg(ap, long *);
            parse_long(*argv++, lp);
            break;
        case 'f':
            fp = va_arg(ap, float *);
            parse_float(*argv++, fp);
            break;
        case 'd':
            dp = va_arg(ap, double *);
            parse_double(*argv++, dp);
            break;
	case 's':
	    sp = va_arg(ap, const char**);
	    *sp = *argv++;
	    break;
        case '\0':
	    usage_error("Too many <perf_args>.");
            break;
        default:
            error("Format string contains illegal characters.");
            break;
        }
    }

    va_end(ap);

    arena_init();

    find_timer_overhead();
}
